import torch
from torch import nn, einsum
import torch.nn.functional as F

from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from ltr import model_constructor
from util import box_ops


from ltr.models.loss.matcher import build_matcher
from ltr.models.backbone.swin_transformer import  build_swinT_base_model_sm, build_swinT_base_model_sm_m, \
    build_swinT_base_model_sm_m_1k, build_swinT_base_model_sm_1k

from ltr.models.head.mask import Mask_Predictor_fine, Mask_Predictor_mask
# helpers

def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d

# pre-layernorm

class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn
    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)

# feedforward

class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout = 0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout)
        )
    def forward(self, x):
        return self.net(x)

# attention

class Attention(nn.Module):
    def __init__(self, dim, heads = 8, dim_head = 64, dropout = 0.):
        super().__init__()
        inner_dim = dim_head *  heads
        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim = -1)
        self.to_q = nn.Linear(dim, inner_dim, bias = False)
        self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x, context = None, kv_include_self = False):
        b, n, _, h = *x.shape, self.heads
        context = default(context, x)

        if kv_include_self:
            context = torch.cat((x, context), dim = 1) # cross attention requires CLS token includes itself as key / value

        qkv = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), qkv)

        dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale

        attn = self.attend(dots)

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)

# transformer encoder, for small and large patches

class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout = 0.):
        super().__init__()
        self.layers = nn.ModuleList([])
        self.norm = nn.LayerNorm(dim)
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, heads = heads, dim_head = dim_head, dropout = dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, dropout = dropout))
            ]))

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return self.norm(x)

# projecting CLS tokens, in the case that small and large patch tokens have different dimensions

class ProjectInOut(nn.Module):
    def __init__(self, dim_in, dim_out, fn):
        super().__init__()
        self.fn = fn

        need_projection = dim_in != dim_out
        self.project_in = nn.Linear(dim_in, dim_out) if need_projection else nn.Identity()
        self.project_out = nn.Linear(dim_out, dim_in) if need_projection else nn.Identity()

    def forward(self, x, *args, **kwargs):
        x = self.project_in(x)
        x = self.fn(x, *args, **kwargs)
        x = self.project_out(x)
        return x

# cross attention transformer

class CrossTransformer(nn.Module):
    def __init__(self, sm_dim, lg_dim, depth, heads, dim_head, dropout):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                ProjectInOut(sm_dim, lg_dim, PreNorm(lg_dim, Attention(lg_dim, heads = heads, dim_head = dim_head, dropout = dropout))),
                ProjectInOut(lg_dim, sm_dim, PreNorm(sm_dim, Attention(sm_dim, heads = heads, dim_head = dim_head, dropout = dropout)))
            ]))

    def forward(self, sm_tokens, lg_tokens):
        #(sm_cls, sm_patch_tokens), (lg_cls, lg_patch_tokens) = map(lambda t: (t[:, :1], t[:, 1:]), (sm_tokens, lg_tokens))

        for sm_attend_lg, lg_attend_sm in self.layers:
            #sm_cls = sm_attend_lg(sm_cls, context = lg_patch_tokens, kv_include_self = True) + sm_cls
            #lg_cls = lg_attend_sm(lg_cls, context = sm_patch_tokens, kv_include_self = True) + lg_cls

            sm_tokens = sm_attend_lg(sm_tokens, context=lg_tokens, kv_include_self=True) + sm_tokens
            lg_tokens = lg_attend_sm(lg_tokens, context=sm_tokens, kv_include_self=True) + lg_tokens

        #sm_tokens = torch.cat((sm_cls, sm_patch_tokens), dim = 1)
        #lg_tokens = torch.cat((lg_cls, lg_patch_tokens), dim = 1)
        return sm_tokens, lg_tokens

class CrossTransformer_fusion(nn.Module):
    def __init__(self, sm_dim, lg_dim, depth, heads, dim_head, dropout):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                ProjectInOut(lg_dim, sm_dim, PreNorm(sm_dim, Attention(sm_dim, heads = heads, dim_head = dim_head, dropout = dropout)))
            )

    def forward(self, sm_tokens, lg_tokens):

        for lg_attend_sm in self.layers:
            lg_tokens = lg_attend_sm(lg_tokens, context = sm_tokens, kv_include_self = True) + lg_tokens

        return lg_tokens

class MultiScaleEncoder(nn.Module):
    def __init__(
        self,
        *,
        depth,
        sm_dim,
        lg_dim,
        sm_enc_params,
        lg_enc_params,
        cross_attn_heads,
        cross_attn_depth,
        cross_attn_dim_head = 64,
        dropout = 0.
    ):
        super().__init__()
        self.layers = nn.ModuleList([])
        self.layers_fusion = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                CrossTransformer(sm_dim=sm_dim, lg_dim=lg_dim, depth = cross_attn_depth, heads = cross_attn_heads, dim_head = cross_attn_dim_head, dropout = dropout)
            ]))

        self.layers_fusion.append(nn.ModuleList([
            CrossTransformer_fusion(sm_dim=sm_dim, lg_dim=lg_dim, depth=1, heads=cross_attn_heads,
                             dim_head=cross_attn_dim_head, dropout=dropout)
        ]))

    def forward(self, sm_tokens, lg_tokens):
        for cross_attend in self.layers:
            #sm_tokens, lg_tokens = sm_enc(sm_tokens), lg_enc(lg_tokens)
            sm_tokens, lg_tokens = cross_attend[0](sm_tokens, lg_tokens)
        for cross_attend_fuse in self.layers_fusion:
            lg_tokens_fusion = cross_attend_fuse[0](sm_tokens, lg_tokens)
        return sm_tokens, lg_tokens, lg_tokens_fusion

# patch-based image to token embedder

class ImageEmbedder(nn.Module):
    def __init__(
        self,
        *,
        dim,
        image_size,
        patch_size,
        dropout = 0.
    ):
        super().__init__()
        assert image_size % patch_size == 0, 'Image dimensions must be divisible by the patch size.'
        num_patches = (image_size // patch_size) ** 2
        patch_dim = 3 * patch_size ** 2

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = patch_size, p2 = patch_size),
            nn.Linear(patch_dim, dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.dropout = nn.Dropout(dropout)

    def forward(self, img):  # img [1,3,256,256]
        x = self.to_patch_embedding(img)  # [1, 256, 192]
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, '() n d -> b n d', b = b) # [1,1,192]
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, :(n + 1)]

        return self.dropout(x)

# cross ViT class

class CrossViT(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        #self.SwinTransformer_lg = build_swinT_base_model_lg()
        self.SwinTransformer_sm = build_swinT_base_model_sm_1k()

        self.multi_scale_encoder = MultiScaleEncoder(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l,img_s):

        sm_tokens_lg = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]

        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)
        lg_tokens = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]
        """"""
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)
        """"""

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}
        return out

    def track(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, features_search)

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}

        return out

    def track_st(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf.clone()

    def instance(self, x):
        with torch.no_grad():
           xf= self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf = self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

class CrossViT_1(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        #self.SwinTransformer_lg = build_swinT_base_model_lg()
        self.SwinTransformer_sm = build_swinT_base_model_sm()

        """
            SwinTransformer(img_size=224,
                                patch_size=4,
                                in_chans=3,
                                num_classes=1000,
                                embed_dim=128,
                                depths=[2, 2, 18, 2],
                                num_heads=[4, 8, 16, 32],
                                window_size=7,
                                mlp_ratio=4.,
                                qkv_bias=True,
                                qk_scale=None,
                                drop_rate=0,
                                drop_path_rate= 0.2,
                                ape=False,
                                patch_norm=True,
                                use_checkpoint=False)
        """
        #self.lg_image_embedder = ImageEmbedder(dim = lg_dim, image_size = image_size, patch_size = lg_patch_size, dropout = emb_dropout)

        self.multi_scale_encoder = MultiScaleEncoder(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        #self.sm_mlp_head = nn.Sequential(nn.LayerNorm(sm_dim), nn.Linear(sm_dim, num_classes))
        #self.lg_mlp_head = nn.Sequential(nn.LayerNorm(lg_dim), nn.Linear(lg_dim, num_classes))
        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l,img_s, img_dym=None):

        sm_tokens_lg = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]

        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)
        lg_tokens = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}
        return out

    def track(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}

        return out

    def track_st(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf

    def instance(self, x):
        with torch.no_grad():
           xf= self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf = self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

class MultiScaleEncoder_dym(nn.Module):
    def __init__(
        self,
        *,
        depth,
        sm_dim,
        lg_dim,
        sm_enc_params,
        lg_enc_params,
        cross_attn_heads,
        cross_attn_depth,
        cross_attn_dim_head = 64,
        dropout = 0.
    ):
        super().__init__()
        self.layers = nn.ModuleList([])
        self.layers_fusion = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                CrossTransformer(sm_dim=sm_dim, lg_dim=lg_dim, depth = cross_attn_depth, heads = cross_attn_heads, dim_head = cross_attn_dim_head, dropout = dropout)
            ]))

        self.layers_fusion.append(nn.ModuleList([
            CrossTransformer_fusion(sm_dim=sm_dim, lg_dim=lg_dim, depth=1, heads=cross_attn_heads,
                             dim_head=cross_attn_dim_head, dropout=dropout)
        ]))

    def forward(self, sm_tokens, sm_tokens_dym,  lg_tokens):
        sm_tokens = torch.cat([sm_tokens, sm_tokens_dym], dim=1)
        for cross_attend in self.layers:
            #sm_tokens, lg_tokens = sm_enc(sm_tokens), lg_enc(lg_tokens)
            sm_tokens, lg_tokens = cross_attend[0](sm_tokens, lg_tokens)
        for cross_attend_fuse in self.layers_fusion:
            lg_tokens_fusion = cross_attend_fuse[0](sm_tokens, lg_tokens)
        return sm_tokens, lg_tokens, lg_tokens_fusion

class CrossViTseg(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        self.SwinTransformer_sm = build_swinT_base_model_sm_m()
        self.multi_scale_encoder = MultiScaleEncoder(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)
        self.seg_network = Mask_Predictor_mask()


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l, img_s, img_s_dym=None):

        sm_tokens_lg, _ = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)


        lg_tokens, inter_out = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        #### seg branch ####

        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict
               }
        return out


    def track_segm(self, search):

        features_search, inter_out = self.SwinTransformer_sm(search)  # [1, 14*14, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, self.dym_temp, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict}
        return out

    def track_st(self, search):

        features_search, _ = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st, _ = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def dynamic_temp(self, z_st):
        with torch.no_grad():
            zf_st, _ = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]
            dym_temp = self.crop_z_feature(zf_st, b, hw, c)
            self.dym_temp = dym_temp

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg, _ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf

    def instance(self, x):
        with torch.no_grad():
           xf, _ = self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf , _= self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg, _ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf, _ = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

class CrossViTseg_2(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        self.SwinTransformer_sm = build_swinT_base_model_sm_m_1k()
        self.multi_scale_encoder = MultiScaleEncoder_dym(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        #self.sm_mlp_head = nn.Sequential(nn.LayerNorm(sm_dim), nn.Linear(sm_dim, num_classes))
        #self.lg_mlp_head = nn.Sequential(nn.LayerNorm(lg_dim), nn.Linear(lg_dim, num_classes))
        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)
        self.seg_network = Mask_Predictor_mask()


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l, img_s, img_s_dym):

        sm_tokens_lg, _ = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)

        sm_tokens_lg_dym, _ = self.SwinTransformer_sm(img_s_dym)  # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg_dym.shape[0], sm_tokens_lg_dym.shape[1], sm_tokens_lg_dym.shape[2]
        sm_tokens_dym = self.crop_z_feature(sm_tokens_lg_dym, b, hw, c)

        lg_tokens, inter_out = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, sm_tokens_dym, lg_tokens)

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        #### seg branch ####

        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict
               }
        return out


    def track_segm(self, search):

        features_search, inter_out = self.SwinTransformer_sm(search)  # [1, 14*14, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, self.dym_temp, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict}
        return out

    def track_st(self, search):

        features_search, _ = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st, _ = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def dynamic_temp(self, z_st):
        with torch.no_grad():
            zf_st, _ = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]
            dym_temp = self.crop_z_feature(zf_st, b, hw, c)
            self.dym_temp = dym_temp

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg, _ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf

    def instance(self, x):
        with torch.no_grad():
           xf, _ = self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf , _= self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg, _ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf, _ = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

class CrossViTdym(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        #self.SwinTransformer_lg = build_swinT_base_model_lg()
        self.SwinTransformer_sm = build_swinT_base_model_sm_m()

        self.multi_scale_encoder = MultiScaleEncoder_dym(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        #self.sm_mlp_head = nn.Sequential(nn.LayerNorm(sm_dim), nn.Linear(sm_dim, num_classes))
        #self.lg_mlp_head = nn.Sequential(nn.LayerNorm(lg_dim), nn.Linear(lg_dim, num_classes))
        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)

        self.avg = torch.nn.AdaptiveAvgPool1d(1)

        self.seg_network = Mask_Predictor_mask()



    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l,img_s, img_s_dym):

        sm_tokens_lg, _ = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)

        lg_tokens, inter_out = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]

        sm_tokens_lg_dym, _ = self.SwinTransformer_sm(img_s_dym)  # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg_dym.shape[0], sm_tokens_lg_dym.shape[1], sm_tokens_lg_dym.shape[2]
        sm_tokens_dym = self.crop_z_feature(sm_tokens_lg_dym, b, hw, c)

        sm_tokens_dym = sm_tokens_dym.permute(0, 2, 1)
        sm_tokens_dym = self.avg(sm_tokens_dym).permute(0, 2, 1)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, sm_tokens_dym, lg_tokens)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        #### seg branch ####
        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict,
               }
        return out

    def track(self, search):

        features_search, inter_out = self.SwinTransformer_sm(search)  # [1, 14*14, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, self.dym_temp, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  # [1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        # mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}
        return out

    def track_st(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg,_ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf.clone()

    def instance(self, x):
        with torch.no_grad():
           xf= self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf = self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def track_segm(self, search):

        features_search, inter_out = self.SwinTransformer_sm(search)  # [1, 14*14, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, self.dym_temp, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        #mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}
        return out

    def dynamic_temp(self, z_st):
        with torch.no_grad():
            zf_st, _ = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]
            dym_temp = self.crop_z_feature(zf_st, b, hw, c)
            dym_temp = dym_temp.permute(0, 2, 1)
            avg = torch.nn.AdaptiveAvgPool1d(1)
            self.dym_temp = avg(dym_temp).permute(0, 2, 1)

class CrossViT_context(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        self.SwinTransformer_sm = build_swinT_base_model_sm()
        self.multi_scale_encoder = MultiScaleEncoder(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l,img_s):

        sm_tokens_lg = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]

        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)
        lg_tokens = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]
        """"""

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)
        """"""

        #sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}
        return out

    def track(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, features_search)

        #sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1]}

        return out

    def track_st(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf.clone()

    def instance(self, x):
        with torch.no_grad():
           xf= self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf = self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf

    def track_motion(self, x, zf):
        xf = self.SwinTransformer_sm(x)

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        xs = self.SwinTransformer_sm.forward_stride8(x)
        bz, wh, c = xs.shape
        h = int(wh ** 0.5)
        xs_square = xs.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################

        return {
            'score': score,
            'bbox':pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

    def tracking_motion(self, xf, zf):
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf, xf)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        bz, wh, c = xf.shape
        h = int(wh ** 0.5)
        xs_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        score = self._convert_score(outputs_class[-1])  ##################
        pred_bbox = self._convert_bbox(outputs_coord[-1]) #######################
        return {
            'score': score,
            'bbox': pred_bbox,
            'feature': xs_square,
            'search_feat': xf
        }

class CrossViT_22k(nn.Module):
    def __init__(
        self,
        *,
        sm_dim,
        lg_dim,
        sm_patch_size = 12,
        sm_enc_depth = 1,
        sm_enc_heads = 8,
        sm_enc_mlp_dim = 2048,
        sm_enc_dim_head = 64,
        lg_patch_size = 16,
        lg_enc_depth = 4,
        lg_enc_heads = 8,
        lg_enc_mlp_dim = 2048,
        lg_enc_dim_head = 64,
        cross_attn_depth = 2,
        cross_attn_heads = 8,
        cross_attn_dim_head = 64,
        depth = 3,
        dropout = 0.1,
        emb_dropout = 0.1
    ):
        super().__init__()
        #self.SwinTransformer_lg = build_swinT_base_model_lg()
        self.SwinTransformer_sm = build_swinT_base_model_sm_m()

        self.multi_scale_encoder = MultiScaleEncoder(
            depth = depth,
            sm_dim = sm_dim,
            lg_dim = lg_dim,
            cross_attn_heads = cross_attn_heads,
            cross_attn_dim_head = cross_attn_dim_head,
            cross_attn_depth = cross_attn_depth,
            sm_enc_params = dict(
                depth = sm_enc_depth,
                heads = sm_enc_heads,
                mlp_dim = sm_enc_mlp_dim,
                dim_head = sm_enc_dim_head
            ),
            lg_enc_params = dict(
                depth = lg_enc_depth,
                heads = lg_enc_heads,
                mlp_dim = lg_enc_mlp_dim,
                dim_head = lg_enc_dim_head
            ),
            dropout = dropout
        )

        self.class_embed = MLP(1024, 1024, 1 + 1, 3)
        self.bbox_embed = MLP(1024, 1024, 4, 3)
        self.seg_network = Mask_Predictor_mask()


    def crop_z_feature(self, sm_tokens_lg, b, hw, c):
        h= int(hw**0.5)
        sm_tokens_lg_bhwc = sm_tokens_lg.reshape(b, h, -1, c)
        crop_z_feature = sm_tokens_lg_bhwc[:, 3:10, 3:10, :]
        crop_z_feature = crop_z_feature.reshape(b, -1, c)

        return crop_z_feature

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def forward(self, img_l,img_s):

        sm_tokens_lg, _ = self.SwinTransformer_sm(img_s) # [1, 7*7, 512]
        b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]

        sm_tokens = self.crop_z_feature(sm_tokens_lg, b, hw, c)
        lg_tokens, inter_out = self.SwinTransformer_sm(img_l)  # [1, 14*14, 512]
        """"""
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(sm_tokens, lg_tokens)
        """"""

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 196, 1024]
        lg_track_final_reshape = lg_track_final.permute(1,0,2)# [196=14*14, 1, 512]
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]


        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)

        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict
               }
        return out

    def track(self, search):

        features_search, inter_out = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, features_search)

        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]

        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)

        out = {'pred_logits': outputs_class[-1], 'pred_boxes': outputs_coord[-1], 'mask_predict': mask_predict
               }

        return out

    def segm(self, seg_patch):
        lg_tokens, inter_out = self.SwinTransformer_sm(seg_patch)  # [1, 14*14, 512]
        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(self.zf, lg_tokens)
        mask_predict = self.seg_network(lg_tokens_fusion, inter_out)
        return mask_predict

    def track_st(self, search):

        features_search = self.SwinTransformer_sm(search)  # [1, 16*16, 512]
        if self.zf_st is not None:
            zf_st = self.zf_st
        else:
            zf_st= self.zf

        sm_tokens, lg_tokens, lg_tokens_fusion = self.multi_scale_encoder(zf_st, features_search)
        lg_track_final = torch.cat([lg_tokens, lg_tokens_fusion], dim=2)  # [1, 256, 512+512]
        lg_track_final_reshape = lg_track_final.permute(1, 0, 2)  # [196=14*14, 1, 512]

        # sm_logits = self.sm_mlp_head(sm_cls)
        # lg_logits = self.lg_mlp_head(lg_cls)
        reg_cls_feat = lg_track_final_reshape.unsqueeze(0).transpose(1, 2)  #[1,1,256,1024]
        outputs_class = self.class_embed(reg_cls_feat)  # [1, 1, 256, 2]
        outputs_coord = self.bbox_embed(reg_cls_feat).sigmoid()  # [1, 1, 256, 4]
        out = {'pred_logits_st': outputs_class[-1], 'pred_boxes_st': outputs_coord[-1]}
        return out

    def template_short_term(self, z_st):
        with torch.no_grad():
            zf_st = self.SwinTransformer_sm(z_st)
            b, hw, c = zf_st.shape[0], zf_st.shape[1], zf_st.shape[2]

            zf_st = self.crop_z_feature(zf_st, b, hw, c)
            self.zf_st = zf_st

    def template(self, z):
        with torch.no_grad():
             sm_tokens_lg,_ = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
             b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
             zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
             self.zf = zf
             self.dym_temp = zf.clone()

    def instance(self, x):
        with torch.no_grad():
           xf= self.SwinTransformer_sm.forward_stride8(x)
           bz, wh, c = xf.shape
           h = int(wh ** 0.5)
           xf_square = xf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
           self.cf=xf_square

    def instance_f(self, xf):
        with torch.no_grad():
           self.cf=xf

    def get_feature(self, z):
        with torch.no_grad():
            zf = self.SwinTransformer_sm.forward_stride8(z)
            bz, wh, c = zf.shape
            h = int(wh ** 0.5)
            zf_square = zf.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
            return zf_square

    def get_template(self, z):
        with torch.no_grad():
            sm_tokens_lg = self.SwinTransformer_sm(z)  # [1, 7*7, 512]
            b, hw, c = sm_tokens_lg.shape[0], sm_tokens_lg.shape[1], sm_tokens_lg.shape[2]
            zf = self.crop_z_feature(sm_tokens_lg, b, hw, c)
            return zf


@model_constructor
def transcross_vit(settings):

    model = CrossViT(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    device = torch.device(settings.device)
    model.to(device)
    return model

@model_constructor
def transcross_vit_22k(settings):

    model = CrossViT_22k(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    device = torch.device(settings.device)
    model.to(device)
    return model

@model_constructor
def transcross_vit_context(settings):

    model = CrossViT_context(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    device = torch.device(settings.device)
    model.to(device)
    return model

@model_constructor
def transcross_vit_1(settings):

    model = CrossViT_1(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    device = torch.device(settings.device)
    model.to(device)
    return model

@model_constructor
def transcrossSeg_vit(settings = None):

    model = CrossViTseg(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    if settings is None:
        model.cuda()
    else:
       device = torch.device(settings.device)
       model.to(device)
    return model

@model_constructor
def transcrossSeg_vit_2(settings = None):

    model = CrossViTseg_2(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    if settings is None:
        model.cuda()
    else:
       device = torch.device(settings.device)
       model.to(device)
    return model

@model_constructor
def transcrossDym_vit(settings = None):

    model = CrossViTdym(
        depth=4,  # number of multi-scale encoding blocks
        sm_dim=512,  # 192,            # high res dimension
        sm_patch_size=16,  # high res patch size (should be smaller than lg_patch_size)
        sm_enc_depth=2,  # high res depth
        sm_enc_heads=4,  # 8,        # high res heads
        sm_enc_mlp_dim=1024,  # 2048,   # high res feedforward dimension
        lg_dim=512,  # 384,            # low res dimension
        lg_patch_size=16,  # 64,      # low res patch size
        lg_enc_depth=3,  # low res depth
        lg_enc_heads=4,  # 8,        # low res heads
        lg_enc_mlp_dim=1024,  # 2048,   # low res feedforward dimensions
        cross_attn_depth=2,  # cross attention rounds
        cross_attn_heads=4,  # 8,    # cross attention heads
        dropout=0.1,
        emb_dropout=0.1
)
    if settings is None:
        model.cuda()
    else:
       device = torch.device(settings.device)
       model.to(device)
    return model


class SetCriterion(nn.Module):
    """ This class computes the loss for TransT.
    The process happens in two steps:
        1) we compute assignment between ground truth box and the outputs of the model
        2) we supervise each pair of matched ground-truth / prediction (supervise class and box)
    """
    def __init__(self, num_classes, matcher, weight_dict, eos_coef, losses):
        """ Create the criterion.
        Parameters:
            num_classes: number of object categories, always be 1 for single object tracking.
            matcher: module able to compute a matching between target and proposals
            weight_dict: dict containing as key the names of the losses and as values their relative weight.
            eos_coef: relative classification weight applied to the no-object category
            losses: list of all the losses to be applied. See get_loss for list of available losses.
        """
        super().__init__()
        self.num_classes = num_classes
        self.matcher = matcher
        self.weight_dict = weight_dict
        self.eos_coef = eos_coef
        self.losses = losses
        empty_weight = torch.ones(self.num_classes + 1)
        empty_weight[-1] = self.eos_coef
        self.register_buffer('empty_weight', empty_weight)

    def loss_labels(self, outputs, targets, indices, num_boxes, log=True):
        """Classification loss (NLL)
        targets dicts must contain the key "labels" containing a tensor of dim [nb_target_boxes]
        """
        assert 'pred_logits' in outputs
        src_logits = outputs['pred_logits']

        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.cat([t["labels"][J] for t, (_, J) in zip(targets, indices)])
        target_classes = torch.full(src_logits.shape[:2], self.num_classes,
                                    dtype=torch.int64, device=src_logits.device)
        target_classes[idx] = target_classes_o

        loss_ce = F.cross_entropy(src_logits.transpose(1, 2), target_classes, self.empty_weight)
        losses = {'loss_ce': loss_ce}

        if log:
            # TODO this should probably be a separate loss, not hacked in this one here
            #losses['class_error'] = 100 - accuracy(src_logits[idx], target_classes_o)[0]
            pass

        return losses

    def loss_boxes(self, outputs, targets, indices, num_boxes):
        """Compute the losses related to the bounding boxes, the L1 regression loss and the GIoU loss
           targets dicts must contain the key "boxes" containing a tensor of dim [nb_target_boxes, 4]
           The target boxes are expected in format (center_x, center_y, h, w), normalized by the image size.
        """
        assert 'pred_boxes' in outputs
        idx = self._get_src_permutation_idx(indices)
        src_boxes = outputs['pred_boxes'][idx]
        target_boxes = torch.cat([t['boxes'][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction='none')

        losses = {}
        losses['loss_bbox'] = loss_bbox.sum() / num_boxes
        giou, iou = box_ops.generalized_box_iou(
            box_ops.box_cxcywh_to_xyxy(src_boxes),
            box_ops.box_cxcywh_to_xyxy(target_boxes))
        giou = torch.diag(giou)
        iou = torch.diag(iou)
        loss_giou = 1 - giou
        iou = iou
        losses['loss_giou'] = loss_giou.sum() / num_boxes
        losses['iou'] = iou.sum() / num_boxes
        return losses

    def _get_src_permutation_idx(self, indices):
        # permute predictions following indices
        batch_idx = torch.cat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.cat([src for (src, _) in indices])
        return batch_idx, src_idx

    def _get_tgt_permutation_idx(self, indices):
        # permute targets following indices
        batch_idx = torch.cat([torch.full_like(tgt, i) for i, (_, tgt) in enumerate(indices)])
        tgt_idx = torch.cat([tgt for (_, tgt) in indices])
        return batch_idx, tgt_idx

    def get_loss(self, loss, outputs, targets, indices, num_boxes):
        loss_map = {
            'labels': self.loss_labels,
            'boxes': self.loss_boxes
        }
        assert loss in loss_map, f'do you really want to compute {loss} loss?'
        return loss_map[loss](outputs, targets, indices, num_boxes)

    def forward(self, outputs, targets):
        """ This performs the loss computation.
        Parameters:
             outputs: dict of tensors, see the output specification of the model for the format
             targets: list of dicts, such that len(targets) == batch_size.
                      The expected keys in each dict depends on the losses applied, see each loss' doc
        """
        outputs_without_aux = {k: v for k, v in outputs.items() if k != 'aux_outputs'}

        # Retrieve the matching between the outputs of the last layer and the target
        indices = self.matcher(outputs_without_aux, targets)

        # Compute the average number of target boxes accross all nodes, for normalization purposes
        num_boxes_pos = sum(len(t[0]) for t in indices)

        num_boxes_pos = torch.as_tensor([num_boxes_pos], dtype=torch.float, device=next(iter(outputs.values())).device)

        num_boxes_pos = torch.clamp(num_boxes_pos, min=1).item()

        # Compute all the requested losses
        losses = {}
        for loss in self.losses:
            losses.update(self.get_loss(loss, outputs, targets, indices, num_boxes_pos))

        return losses

class MLP(nn.Module):
    """ Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim]))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x

def transt_loss(settings):
    num_classes = 1
    matcher = build_matcher()
    weight_dict = {'loss_ce': 8.334, 'loss_bbox': 5}  # {'loss_ce': 8.334, 'loss_bbox': 5}
    weight_dict['loss_giou'] = 2  # 2
    losses = ['labels', 'boxes']
    criterion = SetCriterion(num_classes, matcher=matcher, weight_dict=weight_dict,
                             eos_coef=0.0625, losses=losses)
    device = torch.device(settings.device)
    criterion.to(device)
    return criterion

"""
v = CrossViT(
    depth = 4,               # number of multi-scale encoding blocks
    sm_dim = 512, #192,            # high res dimension
    sm_patch_size = 16,      # high res patch size (should be smaller than lg_patch_size)
    sm_enc_depth = 2,        # high res depth
    sm_enc_heads = 4,#8,        # high res heads
    sm_enc_mlp_dim = 1024,#2048,   # high res feedforward dimension
    lg_dim = 512,#384,            # low res dimension
    lg_patch_size = 16,#64,      # low res patch size
    lg_enc_depth = 3,        # low res depth
    lg_enc_heads = 4,#8,        # low res heads
    lg_enc_mlp_dim = 1024, #2048,   # low res feedforward dimensions
    cross_attn_depth = 2,    # cross attention rounds
    cross_attn_heads = 4, #8,    # cross attention heads
    dropout = 0.1,
    emb_dropout = 0.1
)

"""

img_s = torch.randn(2, 3, 128, 128)
img_l = torch.randn(2, 3, 256, 256)

#n_parameters = sum(p.numel() for p in v.parameters() )
#print('number of params:', n_parameters)

#pred = v(img_l, img_s) # (1, 1000)

a=1